#/usr/bin/python

"""
New Interface to CSound

I know, I know, another interface would be terrible. But I think it might be fun
to mess around with the idea.

The idea would be a Python Interface. The python would run in realtime and so
would CSound. Or it could be a design-time thing. You would mess around with the
knobs and such and it would give you a nice techno song.

So the next step?

The tkinter interface is not going to be as easy as I thought sadly...

Song15 has the rest!!! Random sounds good. Check out song15h and song15i.
So how do I correctly add this functionality? It should be limited by the patch
length...

There are a few things I'd like to add:

1) Load Orchestra from file.
	1a) Import Orchestra is easy!
	1b) Importing others orchestras would be useful.
	1c) You can import any instrument using the Instrument subclass system.
2) More Instruments
	X 2a) Inf Instrument is okay if it ever gets a use.
	2b) See what kind of instruments you can grab...
	X 2c) Buzz instrument?
	X 2d) Elegance!
"""

iInstrumentIndex = 1

from random import random

class Instrument:
	"""
	An instrument is a list of instructions that produce a certain type of sound
	given input.
	"""
	def __init__(self, sName='generic instrument', lExtents=None):
		if lExtents == None:
			self.lExtents = [Extent(0, 15, None)]
		else:
			self.lExtents = lExtents
		#end if
		self.iInstrument = iInstrumentIndex
		globals()['iInstrumentIndex'] += 1
		self.sName = sName
	#end def __init__()

	def setExtents(self, lExtents):
		"""
		Allows the user to do randomness very well.
		"""
		if len(lExtents) == len(self.lExtents):
			self.lExtents = lExtents
		else:
			print 'Error setting extents!'
		#end if
	#end def setExtents(lExtents)

	def playNote(self, tStart, tDuration):
		"""
		Creates a line to send to CSound. Does not exist in the instrument,
		but rather in the patch.
		"""
		return [self.iInstrument, tStart, tDuration]
	#end def playNote(tStart, tDuration)

	def playRandomNote(self, tStart):
		"""
		Plays a random note based on extents.
		"""
		lReturn = [self.iInstrument, tStart]
		for oExtent in self.lExtents:
			lReturn.append(oExtent.randomWithin()) # very smooth
		#next oExtent
		return lReturn
	#end def playNote(tStart, tDuration)

	def printOutput(self):
		"""
		Used by Orchestra to make an Orchestra File (*.orc) from instruments.
		"""
		print 'instr', self.iInstrument, ';', self.sName
		print '        out a1'
		print 'endin;', self.sName
		print
	#end def printOutput()

#end class Instrument

class Extent:
	"""
	Extent ensures that an instrument is played within physical limits...
	"""
	def __init__(self, fMin, fMax, fStep):
		self.fMin = fMin
		self.fMax = fMax
		self.fStep = fStep
	#end def __init__()

	def within(self, sValue):
		"""
		Ensure that it is within extents
		"""
		if(self.fMin != None) and (sValue < self.fMin): return False
		if(self.fMax != None) and (sValue > self.fMax): return False
		if(self.fStep != None) and ((sValue - self.fMin) % self.fStep != 0): return False
		return True
	#end def within(sValue)

	def randomWithin(self):
		"""
		Create a number that it is within extents.
		"""
		# Error checking: They need both min and max.
		if self.fMin == None or self.fMax == None:
			return 0.0
		#end if
		# Without step is legal, just different.
		if self.fStep == None or self.fStep == 0:
			fReturn = random() * (self.fMax - self.fMin) + self.fMin
		else:
			# Steps are pretty easy. Requires an integer number of steps.
			# This may fail in very odd instances where there are many steps in
			# the range.
			fSteps = random() * (self.fMax - self.fMin) / self.fStep
			iSteps = int(fSteps)
			fReturn = self.fMin + self.fStep * iSteps
		#end if
		return fReturn
	#end def randomWithin()

#end class Instrument

class Orchestra:
	"""
	An orchestra is made of instruments being played.
	"""
	def __init__(self, iChannels=1):
		"""
		Parse an orchestra file.
		"""
		self.lInstruments = []
		self.iChannels = iChannels
	#end def __init__([sFilename])

	def addInstrument(self, oInstrument):
		"""
		Add an instrument to the orchestra.
		"""
		self.lInstruments.append(oInstrument)
		return oInstrument
	#end def addNote(sNote)

	def printOutput(self):
		"""
		Create a Orchestra File (*.sco) from instruments.
		"""
		print 'sr = 44100' # sound rate
		print 'kr = 4410'  # k rate
		print 'ksmps = 10' # k samples = sr / kr
		print 'nchnls =', self.iChannels  # important because it effects
							# out a1 vs out a1, a1
		print
		for oInstrument in self.lInstruments:
			oInstrument.printOutput()
		#next oInstrument
	#end def printOutput()
#end class Orchestra

class Patch:
	"""
	A patch is made of notes being played.
	"""
	def __init__(self, tStart=0.0, tLength=120.0):
		self.lNotes = []
		self.tStart = tStart
	#end def __init__()

	def addNote(self, lNote):
		"""
		Add a note to the playlist.
		"""
		# error checking is required.
		if lNote != None:
			self.lNotes.append(lNote)
		#end if
	#end def addNote(sNote)

	def printOutput(self):
		"""
		Create a Score File (*.sco) from patches.
		"""
		for lNote in self.lNotes:
			# error checking is required.
			if len(lNote) < 3: continue
			# The first three must be: instrument, start, duration.
			iInstrument	= lNote.pop(0)
			tStart		= lNote.pop(0)
			tDuration	= lNote.pop(0)
			sNote = 'i' + str(iInstrument) + ' ' + sigDigits(tStart + self.tStart) + ' ' + sigDigits(tDuration)
			for sOption in lNote:
				sNote += ' ' + sigDigits(sOption)
			#next sOption
			print sNote
		#next lNote
	#end def printOutput()

#end class Patch

class Score:
	def __init__(self):
		"""
		Parse an score file.
		"""
		self.lPatches = []
		self.lFunctions = []
	#end def __init__([sFilename])

	def addPatch(self, oPatch):
		"""
		Add a note to the playlist.
		"""
		self.lPatches.append(oPatch)
	#end def addPatch(oPatch)

	def addFunction(self, sFunction):
		"""
		Very Important. Function is required for all scores to my limited
		knowledge.
		"""
		self.lFunctions.append(sFunction)
	#end def addPatch(oPatch)

	def printOutput(self):
		"""
		Create a Score File (*.sco) from patches.
		"""
		# TODO: Error checking here

		for sFunction in self.lFunctions:
			print sFunction
		#next oPatch
		print

		for oPatch in self.lPatches:
			oPatch.printOutput()
		#next oPatch
	#end def printOutput()
#end class Score


class Song:
	"""
	The Generic Song is subclassed to create any songs.
	"""
	def __init__(self):
		"""
		Parse a Unified CSound file.
		"""
		self.oOrchestra = None
		self.oScore = None
	#end def __init__([sFilename])

	def addPatch(self, oPatch):
		"""
		Add a note to the playlist.
		"""
		self.lPatches.append(oPatch)
	#end def addPatch(oPatch)

	def printOutput(self):
		"""
		Create a Unified CSound File (*.csd) from Csound.Song.
		"""
		print '<CsoundSynthesizer>'
		print '<CsOptions>'
		# use this if you want it to play...
		# print '-Wdo devaudio'
		print '</CsOptions>'
		print
		print '<CsInstruments>'
		self.oOrchestra.printOutput()
		print '</CsInstruments>'
		print
		print '<CsScore>'
		self.oScore.printOutput()
		print '</CsScore>'
		print
		print '</CsoundSynthesizer>'
	#end def printOutput()
#end class Song


def sigDigits(fValue, iSigDigits=8):
	"""
	Print a number with a certain number of significant digits.
	"""
	sValue = str(fValue)
	if len(sValue) <= iSigDigits:
		return sValue
	#end if
	if '.' in sValue[0:iSigDigits]:
		return sValue[0:iSigDigits]
	else:
		if '.' in sValue:
			return sValue[0:sValue.index('.')]
		else:
			return sValue
		#end if
	#end if
#end def printSigDigits()






